"""Define an 17track.com profile manager."""
import json
from typing import Union

from requests import Session

from py17track.api import BaseAPI
from py17track.exceptions import UnauthenticatedError
from py17track.package import PACKAGE_STATUS_MAP, Package

API_BUYER = 'https://buyer.17track.net/orderapi/call'
API_USER = 'https://user.17track.net/userapi/call'


def requires_authentication(function):
    """Define a check for authentication"""

    def decorator(self, *args, **kwargs):
        """ Decorate! """
        if not self.authenticated:
            raise UnauthenticatedError('You need to authenticate first!')
        else:
            return function(self, *args, **kwargs)

    return decorator


class ProfileManager(BaseAPI):  # pylint: disable=too-few-public-methods
    """Define a 17track.net profile manager."""

    def __init__(self, session: Session) -> None:
        """Initialize."""
        self.authenticated = False
        super().__init__(session)

    def authenticate(self, email: str, password: str) -> None:
        """Authenticate an account."""
        resp = self.post(
            'https://user.17track.net/userapi/call',
            json={
                'version': '1.0',
                'method': 'Signin',
                'param': {
                    'Email': email,
                    'Password': password,
                    'CaptchaCode': ''
                },
                'sourcetype': 0
            })

        if resp.json()['Code'] != 0:
            raise UnauthenticatedError('Invalid username/password')

        self.authenticated = True

    @requires_authentication
    def packages(self,
                 package_state: Union[int, str] = '',
                 show_archived: bool = False) -> list:
        """Get detailed information on packages in an account."""
        resp = self.post(
            API_BUYER,
            json={
                'version': '1.0',
                'method': 'GetTrackInfoList',
                'param': {
                    'IsArchived': show_archived,
                    'Item': '',
                    'Page': 1,
                    'PerPage': 40,
                    'PackageState': package_state,
                    'Sequence': '0'
                },
                'sourcetype': 0
            })

        packages = []  # type: ignore
        for package in resp.json().get('Json', []):
            last_event = package.get('FLastEvent')
            if last_event:
                event = json.loads(last_event)
            else:
                event = {}

            kwargs = {
                'destination_country': package.get('FSecondCountry', 0),
                'info_text': event.get('z'),
                'location': event.get('c'),
                'origin_country': package.get('FFirstCountry', 0),
                'package_type': event.get('d') or 0,
                'status': package.get('FPackageState', 0)
            }
            packages.append(Package(package['FTrackNo'], **kwargs))
        return packages

    @requires_authentication
    def summary(self, show_archived: bool = False) -> dict:
        """Get a quick summary of how many packages are in an account."""
        resp = self.post(
            API_BUYER,
            json={
                'version': '1.0',
                'method': 'GetIndexData',
                'param': {
                    'IsArchived': show_archived
                },
                'sourcetype': 0
            })

        results = {}
        for kind in resp.json().get('Json', {}).get('eitem', []):
            results[PACKAGE_STATUS_MAP[kind['e']]] = kind['ec']
        return results



"""Define a base object for interacting with 17track.net."""

from urllib.parse import urlparse

import requests

from py17track.exceptions import HTTPError

HEADER_ACCEPT = '*/*'
HEADER_USER_AGENT = (
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 '
    '(KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36')


class BaseAPI(object):
    """Define a class that represents an API request."""

    def __init__(self, session=None):
        """Initialize."""
        self.full_url = None
        self.session = session if session else requests.Session()

    def request(self, method_type, url, **kwargs):
        """Define a generic request."""
        url_pieces = urlparse(url)
        host = url_pieces.netloc
        origin = '{0}://{1}'.format(url_pieces.scheme, host)
        referrer = '{0}/en'.format(origin)

        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Accept', HEADER_ACCEPT)
        kwargs['headers'].setdefault('Host', host)
        kwargs['headers'].setdefault('Origin', origin)
        kwargs['headers'].setdefault('Referrer', referrer)
        kwargs['headers'].setdefault('User-Agent', HEADER_USER_AGENT)

        method = getattr(self.session, method_type)
        resp = method(url, **kwargs)

        # I don't think it's good form to make end users of py17track have to
        # explicitly catch exceptions from a sub-library, so here, I wrap the
        # Requests HTTPError in my own:
        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as exc_info:
            raise HTTPError(str(exc_info)) from None

        return resp

    def post(self, url, **kwargs):
        """Define a generic POST request."""
        return self.request('post', url, **kwargs)



"""Define module exceptions."""


class HTTPError(Exception):
    """Define a generic HTTP error (i.e., a wrapper for Requests)."""
    pass


class InvalidTrackingNumberError(ValueError):
    """Define an error for an invalid tracking number."""
    pass


class UnauthenticatedError(Exception):
    """Define an error for attempting to make unauthenticated requests."""
    pass



"""Define an 17track.com client."""

from requests import Session

from .api import BaseAPI
from .profile import ProfileManager
from .track import AdHocTracker


class Client(BaseAPI):  # pylint: disable=too-few-public-methods
    """Define a 17track.net client."""

    def __init__(self) -> None:
        """Initialize."""
        self.session = Session()
        super().__init__(self.session)

        self.profile = ProfileManager(self.session)
        self.track = AdHocTracker(self.session)



"""Define an 17track.com profile manager."""
from requests import Session

from py17track.api import BaseAPI
from py17track.exceptions import InvalidTrackingNumberError
from py17track.package import Package

API_TRACK = 'https://t.17track.net/restapi/track'


class AdHocTracker(BaseAPI):  # pylint: disable=too-few-public-methods
    """Define a 17track.net profile manager."""

    def __init__(self, session: Session) -> None:
        """Initialize."""
        self.authenticated = False
        super().__init__(session)

    def find(self, *tracking_numbers: str) -> list:
        """Get tracking info for one or more tracking numbers."""
        data = {'data': [{'num': num} for num in tracking_numbers]}
        resp = self.post(API_TRACK, json=data)
        json = resp.json()

        if not json.get('dat'):
            raise InvalidTrackingNumberError('Invalid data')

        packages = []
        for info in json['dat']:
            resp_data = info.get('track', {})

            if not resp_data:
                raise InvalidTrackingNumberError()

            kwargs = {
                'destination_country': resp_data.get('c'),
                'info_text': resp_data.get('z0', {}).get('z'),
                'location': resp_data.get('z0', {}).get('c'),
                'origin_country': resp_data.get('b'),
                'package_type': resp_data.get('d', 0),
                'status': resp_data.get('e', 0),
                'tracking_info_language': resp_data.get('ln1', 'Unknown')
            }
            packages.append(Package(info['no'], **kwargs))
        return packages
